from abc import ABCMeta
from functools import wraps
from inspect import Signature, signature
from typing import (  #type: ignore
    Any,
    Callable,
    Dict,
    List,
    Tuple,
    Type,
    _GenericAlias,
)

from ..container.container_instance import ContainerInstance
from ..scope_managers import SingletonScopeManager, TransientScopeManager
from ..validators.decorator_validator import DecoratorValidator
from ..validators.primitive_dependency_validator import PrimitiveDependencyValidator


# Change to implement a ContainerInstance (and then keep Container as the static wrapper)
# UnitTests then only need to be on the Instance implementation
# Test specific methods (clear) can be removed and managed within the UnitTest arrange
class Container:
    base_types: List[Type[Any]] = [type(type), ABCMeta, _GenericAlias]
    container_instance: ContainerInstance = ContainerInstance(
        validators=[PrimitiveDependencyValidator(), DecoratorValidator()],
        scope_managers=[
            TransientScopeManager(base_types=base_types),
            SingletonScopeManager(base_types=base_types),
        ],
        base_types=base_types,
    )

    @staticmethod
    def inject(implementation: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(implementation)
        def wrapper(*args: Tuple[Any], **kwargs: Dict[str, Any]) -> Any:
            if len(kwargs) == 0:
                sig: Signature = signature(implementation)
                if len(sig.parameters) != len(args):
                    # Resolve constructor parameters and append them to kwargs
                    for p in sig.parameters:
                        if p != "self":
                            instance = Container.resolve(
                                interface=sig.parameters[p].annotation
                            )
                            kwargs[p] = instance
            return implementation(*args, **kwargs)

        return wrapper

    @staticmethod
    def add_transient(interface: Type[Any], implementation: Any) -> None:
        Container.container_instance.add_transient(
            interface=interface, implementation=implementation
        )

    @staticmethod
    def add_singleton(interface: Type[Any], implementation: Any) -> None:
        Container.container_instance.add_singleton(
            interface=interface, implementation=implementation
        )

    @staticmethod
    def resolve(interface: Type[Any]) -> Any:
        return Container.container_instance.resolve(interface=interface)

    @staticmethod
    def validate() -> None:
        Container.container_instance.validate()
